/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.valueprovider;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.ETypedElement;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ComboBoxEditingSupport;

/**
 * Base class for {@link IAnnotationValueProvider}s that manages {@link EStructuralFeature}s
 * contained by a {@link IAnnotatedSpecification}.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public abstract class EStructuralFeatureValueProviderBase<T extends IAnnotatedSpecification>
		extends ValueProviderBase<T> {

	/**
	 * Map: Instance key -> {@link EStructuralFeature}s of the {@link IAnnotatedSpecification}
	 * managed by
	 * this {@link IAnnotationValueProvider}.
	 */
	final protected Map<String, EStructuralFeature> structuralFeatureMap;

	/**
	 * Flag if multiple {@link EStructuralFeature}s are managed by this
	 * {@link IAnnotationValueProvider}.
	 */
	protected boolean handlesMultipleEstructuralFeatures;

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a single {@link EStructuralFeature}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EStructuralFeature structuralFeature) {

		super(annotatedSpecificationEClass);

		structuralFeatureMap = new HashMap<String, EStructuralFeature>();
		structuralFeatureMap.put(DEFAULT_KEY, structuralFeature);

		handlesMultipleEstructuralFeatures = false;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a number of {@link EStructuralFeature}s contained in a single
	 * {@link EClass}. The mapping of of instance keys to the corresponding
	 * {@link EStructuralFeature}s is explicitly provided in a {@link Map}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			Map<String, EStructuralFeature> structuralFeatureMap) {

		super(annotatedSpecificationEClass);

		this.structuralFeatureMap = structuralFeatureMap;

		handlesMultipleEstructuralFeatures = true;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented all {@link EStructuralFeature}s defined locally in a single
	 * {@link EClass} (i.e., not considering inherited attributes). The instance keys, and the
	 * mapping to the corresponding {@link EStructuralFeature}s are automatically derived from the
	 * underlying {@link EClass} (instance keys = field names of {@link EStructuralFeature}).
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass) {

		this(annotatedSpecificationEClass, false);
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented all {@link EStructuralFeature}s defined locally in a single
	 * {@link EClass} (i.e., not considering inherited attributes). The instance keys, and the
	 * mapping to the corresponding {@link EStructuralFeature}s are automatically derived from the
	 * underlying {@link EClass} (instance keys = field names of {@link EStructuralFeature}).
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			boolean inheritedFeatures) {

		super(annotatedSpecificationEClass);

		structuralFeatureMap = new HashMap<String, EStructuralFeature>();

		EList<EStructuralFeature> featureList;
		if(inheritedFeatures) {
			featureList = annotatedSpecificationEClass.getEAllStructuralFeatures();
		} else {
			featureList = annotatedSpecificationEClass.getEStructuralFeatures();
		}

		if(featureList.size() == 1) {
			structuralFeatureMap.put(DEFAULT_KEY, featureList.get(0));
		} else {
			for(EStructuralFeature structuralFeature : featureList) {
				structuralFeatureMap.put(structuralFeature.getName(), structuralFeature);
			}
		}

		handlesMultipleEstructuralFeatures = true;
	}

	/**
	 * Returns the {@link EStructuralFeature} for a given {@code instanceKey}, or throws
	 * an exception, if the key is unknown.
	 * 
	 */
	protected EStructuralFeature getEStructuralFeature(String instanceKey) throws Exception {
		EStructuralFeature structuralFeature = structuralFeatureMap.get(instanceKey);
		if(structuralFeature == null) {
			throw new Exception("Instance key " + instanceKey +
					" is unknown for annotation provider " +
					annotatedSpecificationEClass.getInstanceClassName() + ".");
		}

		return structuralFeature;
	}

	/** Returns the {@link EFactory} for the given {@link EStructuralFeature}. */
	private EFactory getEFactory(EStructuralFeature structuralFeature) {
		return structuralFeature.getEType().getEPackage().getEFactoryInstance();
	}

	/**
	 * Sets a value for a {@link IAnnotatedSpecification} from a {@link String} representation of
	 * the input.
	 */
	private <V> void
			setAnnotationValueFromString(String value, T specification, String instanceKey)
					throws Exception {

		EStructuralFeature structuralFeature = getEStructuralFeature(instanceKey);
		specification.eSet(
				structuralFeature,
				getEFactory(structuralFeature).createFromString(
						((EAttribute)structuralFeature).getEAttributeType(), value));
	}

	/** {@inheritDoc} */
	@Override
	public boolean canEdit(T specification, String instanceKey) {
		return true;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception {
		if(value instanceof String) {
			// In case, the input choice is presented in dynamically managed combo box (i.e., where
			// new value choices can be entered by the user, the empty String indicates that the
			// current value should be deleted, and a neighboring value from the set should be used.
			// In case the last occurrence of a value choice has been removed, it will be no longer
			// be offered in the combo box.
			if(((String)value).isEmpty()) {
				String rval =
						getAnnotationValueAndUpdateInputChoice((String)value, specification,
								instanceKey);
				if(rval != null) {
					value = (U)rval;
				}
			}

			setAnnotationValueFromString((String)value, specification, instanceKey);
		} else {
			specification.eSet(getEStructuralFeature(instanceKey), value);
		}
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		setAnnotationValue(value, specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception {
		// See comment in setAnnotationValue(U value, T specification)
		if((value == null || value.isEmpty())) {
			String rval = getAnnotationValueAndUpdateInputChoice(value, specification, instanceKey);
			if(rval != null) {
				value = rval;
			}
		}

		EStructuralFeature structuralFeature = getEStructuralFeature(instanceKey);

		if(structuralFeature instanceof EAttribute) {
			setAnnotationValueFromString(value, specification, instanceKey);
		} else {
			throw new Exception(
					"setAnnotationValue(String, T) is not supported / has not been implemented the annotation type " +
							structuralFeature.getEType().getName() +
							"(specification class: " +
							specification.getClass() + ").");
		}
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		setAnnotationValue(value, specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception {
		return (U)specification.eGet(getEStructuralFeature(instanceKey));
	}

	/** {@inheritDoc} */
	@Override
	public <U> U getAnnotationValue(T specification) throws Exception {
		return getAnnotationValue(specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@Override
	public EditingSupport createEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, String instanceKey) throws Exception {

		EStructuralFeature structuralFeature = getEStructuralFeature(instanceKey);
		if(structuralFeature.getUpperBound() > 1 ||
				structuralFeature.getUpperBound() == ETypedElement.UNBOUNDED_MULTIPLICITY) {
			throw new Exception(
					"EStructuralValueProvider has not been implemented for attribute multiplicity > 1.");
		}
		EClassifier eType = null;
		try {
			eType = structuralFeature.getEType();
		} catch(Exception e) {
			// Ignore
		}
		if(eType instanceof EEnum) {
			List<String> enumValues = new ArrayList<String>();
			for(EEnumLiteral e : ((EEnum)eType).getELiterals()) {
				enumValues.add(e.getName());
			}

			return new ComboBoxEditingSupport(viewer, clazz, enumValues, instanceKey, false);
		}

		return super.createEditingSupport(viewer, clazz, instanceKey);
	}

	/**
	 * Creates a new instance of the {@link EStructuralFeature} (use to populate newly created
	 * annotation instances created in
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 */
	protected Object createStructuralFeatureInstance(String instanceKey) throws Exception {
		EStructuralFeature structuralFeature = getEStructuralFeature(instanceKey);

		EFactory factory = getEFactory(structuralFeature);
		if(structuralFeature instanceof EAttribute) {

			EDataType attributeEDataType = ((EAttribute)structuralFeature).getEAttributeType();

			if(structuralFeature.getDefaultValueLiteral() != null) {
				// Use default value if it has been specified
				return factory.createFromString(attributeEDataType,
						structuralFeature.getDefaultValueLiteral());
			}

			// Return data types default value
			return attributeEDataType.getDefaultValue();
		} else if(structuralFeature instanceof EReference) {
			EReference reference = (EReference)structuralFeature;
			if(reference.isContainment()) {
				// EClasses attached to contained EReferences are created lazily by the
				// corresponding generated code on the first access (i.e., also during get / "read")
				return null;
			}
			return factory.create(reference.getEReferenceType());
		} else {
			throw new Exception(
					"createStructuralFeatureInstance() is not supported / has not been implemented the annotation type " +
							structuralFeature.getName() + ".");
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void decorateAnnotationSpecification(T specification, String instanceKey)
			throws Exception {

		// Create and set structural feature implementing the annotation
		Object structuralFeatureVal = createStructuralFeatureInstance(instanceKey);
		if(structuralFeatureVal != null) {
			specification.eSet(getEStructuralFeature(instanceKey), structuralFeatureVal);
		}
	}

	/** {@inheritDoc} */
	@Override
	public Collection<String> getInstanceKeys(T specification) {
		if(handlesMultipleEstructuralFeatures) {
			return structuralFeatureMap.keySet();
		}
		return super.getInstanceKeys(specification);
	}
}
